package folio;

import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.ElementStack;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.SAXParser;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class XMLZones implements SAXParserHandler {

    private Locator locator;
    private String school;
    private String filename;
    private String shortname;
    private String imageDirectory;
    private ManuscriptInfo mi;
    private SAXParser saxParser;
    private StringBuilder buffer = new StringBuilder();
    //private int currentZoneNumber = 0;
    private ElementStack stack = new ElementStack();
    private SurfaceInfo currentSurface = null;
    private ZoneInfo currentZone = null;
    private GraphicInfo currentGraphic = null;
    private boolean appendBuffer = false;
    private int currentFileNumber = -1;
    private Hashtable<String, SurfaceInfo> zoneDimensions = new Hashtable<String, SurfaceInfo>();
    private BufferedWriter repairScript;

    public XMLZones(BufferedWriter repairScript) {
        this.repairScript = repairScript;
        saxParser = new SAXParser(this);
    }

    public void parse(String school, String tempfile, String realname, String imageDirectory, ManuscriptInfo mi) throws Exception {
        this.school = school;
        this.filename = realname;
        this.imageDirectory = imageDirectory;
        currentFileNumber = -1;
        mi.surfaces.clear();
        zoneDimensions.clear();
        this.mi = mi;
        buffer.setLength(0);
        shortname = new File(filename).getName();
        if (shortname.equals("09.34_item_33_TIFFcrops_tei.xml")) {
            shortname = "09.34_item_33_TIFFcrops_tei.xml";
        }
        try {
            saxParser.parse(tempfile);
        } catch (Exception ex) {
            Main.writeError(String.format("%s: %s (in XMLZones.java)", realname, ex.getMessage()));
        }
    }

    public void warning(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void error(SAXParseException exception) throws SAXException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void fatalError(SAXParseException exception) throws SAXException {
        String msg = String.format("%s:%d.%d %s (in XMLZones.java)",
                    filename, locator.getLineNumber(), locator.getColumnNumber(), exception.getMessage());
        Main.writeError(msg);
        stack.clear();
        throw new SAXException("Boom!");
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    public void startDocument() throws SAXException {
    }

    public void endDocument() throws SAXException {

        Main.writeError(String.format("\n\n******************\nProcessing %s\n******************\n", mi.directoryName));

        if (mi.folioItems == null) {
            Main.writeError(String.format("%s/%s has no foliation formula.\n", mi.directoryName, mi.catalogName));
        }

        for (SurfaceInfo si : mi.surfaces) {
            zoneDimensions.put(si.graphic.url, si);
        }

        // print error messages
        if (mi.surfaces.size() != mi.imageFiles.size()) {
            if (mi.imageFiles.size() == 0) {
                return;
            }
            
            Hashtable<String, Boolean> h = new Hashtable<String, Boolean>();
            int imageID = 1;
            for (FileInfo fi : mi.imageFiles) {
                if (fi.name.equals("noImage.jpg")) {
                    h.put(String.format("noImage%d.jpg.", imageID), Boolean.TRUE);
                    imageID++;
                } else {
                    h.put(fi.name, Boolean.TRUE);
                }
            }

            imageID = 1;
            Enumeration<String> en = h.keys();
            while (en.hasMoreElements()) {
                String key = en.nextElement();
                if (zoneDimensions.get(key) == null) {
                    if (key.contains("noImage")) {
                        SurfaceInfo si = SurfaceInfo.blankSurface(key);
                        zoneDimensions.put(key, si);
                        mi.surfaces.add(si);
                    } else {
                        Main.writeError(String.format("\t%s is in pages.xml but is not a surface\n", key));
                    }
                }
            }

            if (mi.surfaces.size() != mi.imageFiles.size()) {               // gaps were the only problem this will be false
                Main.writeError(String.format("\n%s has %d surfaces and %d jpg files\n",
                    shortname, mi.surfaces.size(), mi.imageFiles.size()));
            }

            en = zoneDimensions.keys();
            while (en.hasMoreElements()) {
                String key = en.nextElement();
                if (h.get(key) == null) {
                    Main.writeError(String.format("\tsurface %s is not in pages.xml\n", key));
                }
            }
        }

        for (SurfaceInfo si : mi.surfaces) {
            if (si.missingZones > 0) {
                Main.writeError(String.format("Donglei file %s has missing zones\n", shortname));
                break;
            }
        }

        for (SurfaceInfo si : mi.surfaces) {
            if (si.extraZones > 0) {
                Main.writeError(String.format("Donglei file %s has extra zones\n", shortname));
                break;
            }
        }

        for (SurfaceInfo si : mi.surfaces) {
            if (si.addedBy != null) {
                Main.writeError(String.format("Donglei file %s has at least one missing surface added by %s\n",
                        shortname, si.addedBy));
                break;
            }
        }
        File dir = new File(String.format("%s%s/", Main.dirXMLZonesRepaired, school));
        dir.mkdirs();

        String out = String.format("%s/%s", dir.getAbsolutePath(), shortname);
        if (shortname.contains("2174")) {
            shortname = shortname + "";
        }
        try {
            BufferedWriter bw = IOAssistance.openWriter(out);
            bw.write("<?xml version='1.0' encoding='UTF-8' ?>\n");
            bw.write("<zones>\n");
            // Write surfaces in same order as the files in pages.xml
            int pageNumber = 0;
            ArrayList<String> imagePageIDs = new ArrayList<String>();
            int totalPages = 0;
            boolean errorReported = false;
            for (FileInfo fi : mi.imageFiles) {
                imagePageIDs.clear();
                totalPages += fi.numFolioPages;
                for (int i=0; i<fi.numFolioPages; i++) {
                    if (pageNumber >= mi.folioIDs.size()) {
                        imagePageIDs.add("f. ???");
                    } else {
                        imagePageIDs.add(String.format("f. %s", mi.folioIDs.get(pageNumber)));
                    }
                    pageNumber++;
                }
                SurfaceInfo si = zoneDimensions.get(fi.name);
                if (si != null) {               // if null then file in page.xml has not been cropped.
                    si.fixup();
                    if (si.zones.size() != imagePageIDs.size()) {
                        Main.writeError(String.format("%s/%s has %s zones and %d pages.\n", shortname, fi.name, si.zones.size(), imagePageIDs.size()));
                    }
                    si.graphic.desc = join(imagePageIDs, ", ");
                    si.facsGrp = fi.facsGrp;            // if fi.facsGrp is not null, then surface is probably blank
                    fixupZones(si.zones, imagePageIDs, fi.facsid);
                    si.createIDFromZoneIDs(fi.facsid);
                    SurfaceInfo.indent = "  ";
                    bw.write(si.toString());
                    bw.write("\n");
                    mi.surfacesRepaired.add(si);
                } else if (!errorReported && fi.name.contains("2334")) {
                    System.out.format("!!!%s: %s not found\n", shortname, fi.name);
                    Enumeration<String> en = zoneDimensions.keys();
                    while (en.hasMoreElements()) {
                        System.out.format("\t%s\n", en.nextElement());
                    }
                    errorReported = true;
                }
            }
            bw.write("</zones>\n");
            bw.close();
            if (totalPages != mi.folioIDs.size()) {
                /* Already reported elsewhere
                Main.writeError(String.format("%s: foliation formula indicates there should be %d pages, but info in pages.xml say %d pages.",
                        mi.directoryName, mi.pageIDs.size(), totalPages));
                 */
            }
        } catch (Exception ex) {
            Main.writeError(String.format("%s: %s (in XMLZones.java)", shortname, ex.getMessage()));
            throw new SAXException(ex.getMessage());
        }
    }
    
    private String join (ArrayList<String> a, String delim) {
        StringBuilder sb = new StringBuilder();
        String _delim = "";
        for (String p : a) {
            sb.append(_delim);
            sb.append(p);
            _delim = delim;
        }
        return sb.toString();
    }


    // Correct Donglei's folio guesses
    private void fixupZones(ArrayList<ZoneInfo> zones, ArrayList<String> pageIDs, String facsid) {
        int n = 0;
        int numPages = pageIDs.size();
        for (ZoneInfo zi : zones) {
            if (n < numPages) {
                zi.graphic.desc = pageIDs.get(n);
                zi.mapDescToId();
                n++;
            } else if (numPages == 0 && facsid != null) {
                zi.graphic.desc = facsid;
                zi.graphic.setID(facsid);
            } else {
                zi.graphic.desc = "???";
                zi.graphic.setID("???");
            }
        }
    }


    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    public void endPrefixMapping(String prefix) throws SAXException {
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        stack.pushElement(qName, atts);
        String path = stack.getPath();
        try {
            if (path.equals("zones.surface")) {
                currentSurface = new SurfaceInfo();
                String missingZones = atts.getValue("missingZones");
                if (missingZones != null) {
                    currentSurface.missingZones = Integer.parseInt(missingZones);
                }
                String extraZones = atts.getValue("extraZones");
                if (extraZones != null) {
                    currentSurface.extraZones = Integer.parseInt(extraZones);
                }
                currentSurface.addedBy = atts.getValue("addedBy");
                mi.surfaces.add(currentSurface);
                currentFileNumber++;
            } else if (path.equals("zones.surface.zone")) {
                currentZone = new ZoneInfo();
                currentZone.uly = Integer.parseInt(atts.getValue("uly"));
                currentZone.ulx = Integer.parseInt(atts.getValue("ulx"));
                currentZone.lry = Integer.parseInt(atts.getValue("lry"));
                currentZone.lrx = Integer.parseInt(atts.getValue("lrx"));
                currentSurface.zones.add(currentZone);      // in case empty zone added to make things line up
                currentGraphic = new GraphicInfo();
            } else if (path.equals("zones.surface.graphic")) {
                currentGraphic = new GraphicInfo();
                currentGraphic.setID(atts.getValue("xml:id"));
                currentGraphic.url = atts.getValue("url");
                currentSurface.graphic = currentGraphic;
            } else if (path.equals("zones.surface.zone.graphic")) {
                // currentGraphic allocated in zones.surface.zone
                currentGraphic.setID(atts.getValue("xml:id"));
                currentGraphic.url = atts.getValue("url");
                currentZone.graphic = currentGraphic;
            } else if (path.equals("zones.surface.zone.desc")) {
            } else if (path.equals("zones.surface.graphic.desc")) {
            } else if (!path.equals("zones")) {
                Main.writeError(String.format("%s (%d): Unhandled path: %s", filename, locator.getLineNumber(), path));
            }
        } catch (Exception ex) {
            Main.writeError(String.format("%s: %s (in XMLZones.java)", shortname, ex.getMessage()));
        }
    }


    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
            String path = stack.getPath();
            if (path.equals("zones.surface")) {
                if (currentSurface.zones.size() == 2) {
                    ZoneInfo zone1 = currentSurface.zones.get(0);
                    ZoneInfo zone2 = currentSurface.zones.get(1);
                    if (zone1.uly > zone2.uly) {    // pages flipped
                        File oldCrop1 = new File(String.format("%s%s", imageDirectory, zone1.graphic.url));
                        File oldCrop2 = new File(String.format("%s%s", imageDirectory, zone2.graphic.url));
                        String inspect = null;
                        repairScript.write(inspect = String.format("mv %s swap.jpg\n", oldCrop1.getAbsolutePath()));
                        repairScript.write(inspect = String.format("mv %s %s\n",
                                oldCrop2.getAbsolutePath(),
                                oldCrop1.getAbsolutePath()));
                        repairScript.write(inspect = String.format("mv swap.jpg %s\n", oldCrop2.getAbsolutePath()));
                        repairScript.write("\n");
                        
                        GraphicInfo gswap = zone1.graphic;
                        zone1.graphic = zone2.graphic;
                        zone2.graphic = gswap;

                        currentSurface.zones.set(0, zone2);
                        currentSurface.zones.set(1, zone1);
                    }
                }
            } else if (path.equals("zones.surface.graphic.desc")) {     // description of original page
                currentGraphic.desc = "";           // fill this in on endDocument
            } else if (path.equals("zones.surface.zone.desc")) {        // description of split page
                currentGraphic.desc = "";           // fill this in on endDocument
            } else if (path.equals("zones.surface.zone.graphic")) {
                // Do nothing
            } else if (qName.equals("zones")) {
                qName = "zones";
            }

            stack.popElement();
        } catch (Exception ex) {
            Main.writeError(String.format("%s: %s (in XMLZones.java)", shortname, ex.getMessage()));
        }
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        if (appendBuffer) {
            buffer.append(new String(ch, start, length));
        }
    }

    private String fixup(String s) {
        return s;                           // write this if necessary
    }

}
